module prob_nl_l2_diff_mod
! Nonlinear L2 difference objective form
! Last edited by Ali Almomin 08/24/2015
! 0 = Wd (L(S Wm m) - d)

  use prob_obj_mod
  use identity_mod

  implicit none

  private
  public :: prob_nl_l2_diff

  logical, parameter                           :: T = .true., F = .false.

  type, extends(prob_obj)                       :: prob_nl_l2_diff
    private
    integer                           :: nsop
    real, dimension(:), pointer       :: data
    real, dimension(:), allocatable   :: temp_data
    real, dimension(:), allocatable   :: temp_modl
    real, dimension(:), allocatable   :: temp_prec
    procedure(oper), nopass, pointer  :: nlop
    procedure(update), nopass, pointer  :: uop
    procedure(oper), nopass, pointer  :: lop
    procedure(oper), nopass, pointer  :: wdop
    procedure(oper), nopass, pointer  :: wmop
    procedure(oper), nopass, pointer  :: sop
    procedure(oper), nopass, pointer  :: gpop
  contains
    private
    procedure                                  :: objf
    procedure                                  :: resf
    procedure                                  :: gradf
    procedure                                  :: dresf
  end type

  interface prob_nl_l2_diff
    module procedure const
  end interface

  interface
    integer function oper(adj, add, modl, data)
      real,dimension(:) :: modl,data
      logical           :: adj,add
    end function
    integer function update(modl)
      real,dimension(:) :: modl
    end function
  end interface

  contains

  type(prob_nl_l2_diff) function const(nlop, uop, lop, nmodl, ndata, data, wdop, wmop, sop, nsop, gpop)
    optional                     ::                                                        wdop, wmop, sop, nsop, gpop
    procedure(oper)              :: nlop, lop, wdop, wmop, sop, gpop
    procedure(update)            :: uop
    integer                      :: nmodl           ! size of modl (input)
    integer                      :: ndata           ! size of data (input)
    integer                      :: nsop           ! 
    integer                      :: stat           ! 
    real, dimension(:), target   :: data           ! data (input)
    const%nlop => nlop
    const%uop => uop
    const%lop => lop
    const%nmodl = nmodl
    const%ndata = ndata
    const%data => data
    const%wdop => identity_op
    const%wmop => identity_op
    const%sop => identity_op
    const%gpop => identity_op
    const%nsop = nmodl
    if(present(wdop)) const%wdop => wdop
    if(present(wmop)) const%wmop => wmop
    if(present(sop)) const%sop => sop
    if(present(nsop)) const%nsop = nsop
    allocate(const%temp_data(const%ndata))
    allocate(const%temp_modl(const%nmodl))
    allocate(const%temp_prec(const%nsop))
    stat = const%alloc()
  end function

  integer function objf(this, res, obj)        ! function for calculating objective function
    class(prob_nl_l2_diff)        :: this
    real, dimension(:) :: res             ! residual (input)
    double precision   :: obj             ! objective function value (output)
    obj = 0.5*(sum(dprod(res,res)))
    objf = 0
  end function

  integer function resf(this, modl, res)      ! function for calculating the residual
    class(prob_nl_l2_diff)        :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: res             ! residual (output)
    integer                      :: stat            ! status flag
    stat = this%wmop(F, F, modl, this%temp_modl)
    stat = this%sop(F, F, this%temp_modl, this%temp_prec)
    stat = this%nlop(F, F, this%temp_prec, this%temp_data)
    this%temp_data = this%temp_data - this%data
    stat = this%wdop(F, F, this%temp_data, res)
    resf = 0
  end function

  integer function gradf(this, modl, res, grad) ! function for calculating gradient
    class(prob_nl_l2_diff)        :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: res             ! residual (input)
    real, dimension(:) :: grad            ! gradient (output)
    integer                      :: stat   ! status flag
    stat = this%uop(modl)
    stat = this%wdop(T, F, this%temp_data, res)
    stat = this%lop(T, F, this%temp_prec, this%temp_data)
    stat = this%sop(T, F, this%temp_modl, this%temp_prec)
    stat = this%wmop(T, F, this%temp_prec, this%temp_modl)
    stat = this%gpop(T, F, grad, this%temp_prec)
    gradf = 0
  end function

  integer function dresf(this, modl, dmodl, dres) ! function for calculating gradient in data space
    class(prob_nl_l2_diff)        :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: dmodl            ! gradient (input)
    real, dimension(:) :: dres           ! gradient in data space (output)
    integer                      :: stat            ! status flag
    stat = this%uop(modl)
    stat = this%gpop(F, F, dmodl, this%temp_prec)
    stat = this%wmop(F, F, this%temp_prec, this%temp_modl)
    stat = this%sop(F, F, this%temp_modl, this%temp_prec)
    stat = this%lop(F, F, this%temp_prec, this%temp_data)
    stat = this%wdop(F, F, this%temp_data, dres)
    dresf = 0
  end function

end module
